/** -------------------------------------------------------------------
  *      maintainer: Marcus Wolschon <Suran@gmx.net>
  * -------------------------------------------------------------------
  * projects:
  *	- Jaugment II
  *       part: RA-integration
  *	   used in release: current      used in version: current
  * -------------------------------------------------------------------
  * Copyright:   Copyright (c) 2000 by Marcus Wolschon <Suran@gmx.net>
  * -------------------------------------------------------------------
  * changes:
  * 0.0.1 (2000-12-15, Marcus Wolschon<Suran@gmx.net>)
  *       - created
  *       - works
  * 0.0.2 (2000-12-18, Marcus Wolschon<Suran@gmx.net>)
  *       - supports rating
  *       - cuts docid and rating from line
  *       - support multiple indice
  * 0.0.3 (2001-02-09, Marcus Wolschon<Suran@gmx.net>)
  *       - '\n'->' '
  *       - last_question
  * 0.0.4 (2002-02-24, Marcus Wolschon<Suran@gmx.net>)
  *       - stop, restart,... super.* to properly fire events
  * -------------------------------------------------------------------
  * TODO:
  *       - support multiple configurable indexes
  */

package jaugment2.RememberanceAgent;

  //JDK
import java.rmi.RemoteException;
import java.rmi.Remote;
import java.rmi.MarshalledObject;
import java.util.Properties;
import java.util.LinkedList;
import java.util.Hashtable;
import java.util.Set;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.ListIterator;
import java.lang.IllegalArgumentException;
import java.lang.IllegalStateException;
import java.lang.StringBuffer;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.BufferedInputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.util.StringTokenizer;
  //JINI
import com.sun.jini.lease.landlord.Landlord;
import com.sun.jini.lease.landlord.Landlord.RenewResults;
import com.sun.jini.lease.landlord.LocalLandlord;
import com.sun.jini.lease.landlord.LandlordLeaseFactory;
import com.sun.jini.lease.landlord.LandlordUtil;
import com.sun.jini.lease.landlord.LandlordLease;
import com.sun.jini.lease.landlord.LeasedResource;
import com.sun.jini.lease.landlord.LandlordLease.Factory;
import net.jini.core.lease.LeaseDeniedException;
import net.jini.core.lease.UnknownLeaseException;
import net.jini.core.lease.Lease;
import net.jini.core.event.RemoteEvent;
import net.jini.core.event.RemoteEventListener;
import net.jini.core.event.UnknownEventException;
//import net.jini.core.event.EventRegistration;
import net.jini.core.discovery.LookupLocator;
import net.jini.core.entry.Entry;
  //serviceui-draft-standart
import net.jini.lookup.entry.UIDescriptor;
import net.jini.lookup.ui.MainUI;
import net.jini.lookup.ui.factory.JComponentFactory;
import net.jini.lookup.ui.attribute.UIFactoryTypes;
  //project
import jaugment2.jini.simpleService.*;
//import jaugment2.jini.SimpleLandlord;
//import jaugment2.jini.SimpleLandlord.LeaseListener;
//import jaugment2.jini.SimpleLandlord.EventRegistration;
import jaugment2.helpers.Logger;

import jaugment2.RememberanceAgent.RememberanceAgentServiceProxyIF.Answer;

//helper-mark for configure to determine that rmic needs to be executed for this file
//rmimakehelper

/**
 * A service wrapping the MIT-Remenberance-Agent and alowing to
 * integrate it into other services (like the text-editor)
 */
public class RememberanceAgentService extends simpleService implements Remote
{



  //-------------------- simple Service

 /** restart this service once it has been stopped
   * @see stop
   */
 public void restart() throws Exception
 {
  super.restart();
 }

 /** stop this service. If exit is false all references to this
   * service are still valid but it has deregistered everywhere.
   * (this includes the RMI-activation-daemon)
   * @see restart
   * @see destroy
   * @param exit stop running
   */
 public void stop(boolean exit) throws Exception
 {
  super.stop(exit);
 }

 /** stop this service, destroy any persistent storage and exit
   */
 public void destroy() throws RemoteException
 {
  super.destroy();
 }


  //-------------------- Constructor
  public RememberanceAgentService() throws RemoteException
  {
   this(null);
  };


   RememberanceAgentServiceProxy proxy;

  public RememberanceAgentService( LookupLocator firstLocator) throws RemoteException
  {
//   landLord = new SimpleLandlord(simpleLeaselistener);
   try
     {
      persistentState.setProperty("groups", "public, Suran");
      persistentState.setProperty("locators", "jini://localhost, jini://127.0.0.1");

      proxy = new RememberanceAgentServiceProxy(this);


      loadMe();

      Entry[] attrSets = (Entry[])persistentState.get("attr");

      init(proxy, attrSets, this, firstLocator);


//      crypter = new CmdlnCrypter(null/*new CryptixCrypter(null, proxy)*/,proxy);


/*      Answer ans[] = ask("laptop battery mobile ");
      Logger.err.println("got: "+ans.length+" answers:");
      for(int i=0; i<ans.length; i++)
        {
         Logger.err.println(">"+ans[i].toString()+"<");
         if(i==0)
           {
            Logger.err.println("------doc=\n: "+getDocument(ans[i].docid)+"\n------");
           }
        }*/
     }
   catch(Exception e)
     {
      Logger.err.println(e.getMessage());
      e.printStackTrace(Logger.err);
     }

  }

  private void loadMe()
  {
      String filename = System.getProperty("jaugment2.RememberanceAgent.RememberanceAgentService.persistentStateFilename");
      if(filename==null)
        {
         filename = "RememberanceAgentServiceState.dat";
         System.setProperty("jaugment2.RememberanceAgent.RememberanceAgentService.persistentStateFilename", filename);
        }

      persistentState.setProperty("persistentStateFilename", filename);


   load();


   if(persistentState.get("attr")==null)
     {
      Entry[] attrSets = new Entry[]
                    {
                     new net.jini.lookup.entry.Name("RememberanceAgentService"),
                     new net.jini.lookup.entry.Comment("a service integrating the MIT-Rememberance-Agent into jAugement"),
                     new net.jini.lookup.entry.ServiceInfo("RememberanceAgentService", "Marcus Wolschon <Suran@gmx.net>", "Marcus Wolschon <Suran@gmx.net>", "0.0.1", "standart", "00000")//,
//                     PasswordServiceSwingRepresentation.getUIDescriptor(),
//                     PasswordServiceTextRepresentation.getUIDescriptor()
                    };
      persistentState.put("attr", attrSets);
     }

   save();

  }


  private String last_question = "";
  private LinkedList last_retval = new LinkedList();

  //------------------------------------
  public Answer[] ask(String question, int numretvals)
  {
   try
     {
      if(question.equals(last_question))
        return((Answer[])last_retval.toArray(new Answer[0]));

      LinkedList froml = (LinkedList)System.getProperties().get("jaugment2.RememberanceAgent.RememberanceAgentService.indexes");

      LinkedList retval = new LinkedList();

      if(froml == null)
        {
         Logger.err.println("ERROR: [Rememberance-Agent-Service] please set \"jaugment2.RememberanceAgent.RememberanceAgentService.indexes\" to a list of your indice");
        }
      else
      for(int j=0; j<froml.size(); j++)
        {
         String from = froml.get(j).toString();

         LinkedList retval2 =  askme(from, question, numretvals); //TODO: test
         ListIterator r2iter = retval2.listIterator();
         while(r2iter.hasNext())
           {
            Answer a = (Answer)r2iter.next();
            if(retval.size()==0)
              {
               retval.add(a);
               continue;
              };

            for(int i=0; i<(retval.size()+1); i++)
              {
               if(i==numretvals)
                 break;
               if(i==retval.size() || a.rating>(((Answer)retval.get(i)).rating))
                 {
                  retval.add(i, a);
                  if(retval.size()>numretvals)
                    retval.removeLast();
                  break;
                 }
              }
           }
        }

     last_question = question;
     last_retval = retval;
     return((Answer[])retval.toArray(new Answer[0]));
     }
   catch(Exception e)
     {
      e.printStackTrace(Logger.err);
      return(null);
     }
  }




  private LinkedList askme(String base, String question, int numretvals) throws Exception
  {
   LinkedList retval = new LinkedList();
   try
     {
      question = question.replace('\n', ' ');

      Process proc = Runtime.getRuntime().exec("ra-retrieve  "+base);
      OutputStream out = proc.getOutputStream();
      InputStream err = proc.getErrorStream();
      InputStream in = new BufferedInputStream(proc.getInputStream(), 2048*10);
      out.write(("query "+numretvals+"\n"+question+"\n").getBytes());
      out.flush();
      out.close();
      Thread.sleep(1000);
      int aval = in.available();
      for(int i=0; i<5; i++)
      if(aval==0)
        {
         Thread.sleep(1000);
         aval = in.available();
        };
/*     byte retval[] = new byte[aval];

     if(in.read(retval) != aval)
       {Logger.debug.println("DEBUG: [CmddlnCrypter] warning! did not reat specified number of bytes");}
     Logger.err.println("DEBUG: [RememberanceAgentService] I reat "+(new String(retval)));
*/
     aval = err.available();
     if(aval>0)
       {
        byte retvalerr[] = new byte[aval];
        err.read(retvalerr);
        String errmsg = new String(retvalerr);
        if(errmsg.trim().length()>0)
           Logger.err.println("ERROR: [RememberanceAgentService] Maybe I cannot decrypt "+errmsg);
       };


     LineNumberReader lnr = new LineNumberReader(new InputStreamReader(in));
     String line;
     while( (line = lnr.readLine()) != null )
       {
        if(line.equals("."))
          break;
        StringTokenizer t = new StringTokenizer(line, "|");
        try
          {
           String tok = t.nextToken(" ");
           Logger.debug.println("DEBUG: [RememberanceAgentService] tok[0]=\""+tok+"\" ignored");
                  tok = t.nextToken("|");
           Logger.debug.println("DEBUG: [RememberanceAgentService] tok[1]=\""+tok+"\"=rating");
           float rating = Float.parseFloat(tok);
                  tok = t.nextToken();
           Logger.debug.println("DEBUG: [RememberanceAgentService] tok[2]=\""+tok+"\"=docid");
           int docnr = Integer.parseInt(tok.trim());

           String line2 = t.nextToken();
/*           while(t.hasMoreTokens())
             line2 += " | "+t.nextToken();*/


           retval.add(new Answer(docnr, line2, rating, base));
          }
        catch(Exception e)
          {
           Logger.debug.println("DEBUG: [RememberanceAgentService] askme(...) ERROR parsing line=\""+line+"\" command=\""+"query "+numretvals+"\n"+question+"\n"+"\"");
           e.printStackTrace(Logger.debug);
          }
         Logger.debug.println("DEBUG: [RememberanceAgentService] askme(...) line=\""+line+"\" command=\""+"query "+numretvals+"\n"+question+"\n"+"\"");
       }
     return(retval);
    }
  catch(Exception e)
    {
     e.printStackTrace(Logger.err);
    }
   return(new LinkedList());
  }

  public String getDocument(int docid, String base)
  {
   try
     {
      String doc = null;
//      LinkedList froml = (LinkedList)System.getProperties().get("jaugment2.RememberanceAgent.RememberanceAgentService.indexes");
//      for(int j=0; j<froml.size(); j++)
//        {
//         String from = froml.get(j).toString();
         doc =  getDocument2(base, docid);
//         if(doc != null)
//           break;
 //       }
      return(doc);
     }
   catch(Exception e)
     {
      e.printStackTrace(Logger.err);
      return(null);
     }
  }

  private String getDocument2(String base, int docid)
  {
   try
     {
      Process proc = Runtime.getRuntime().exec("ra-retrieve  "+base+" --docnum "+docid);
      InputStream err = proc.getErrorStream();
      InputStream in = new BufferedInputStream(proc.getInputStream(), 2048*10);
      Thread.sleep(1000);
      int aval = in.available();
      for(int i=0; i<5; i++)
      if(aval==0)
        {
         Thread.sleep(1000);
         aval = in.available();
        };
     byte retval[] = new byte[aval];

     if(in.read(retval) != aval)
       {Logger.debug.println("DEBUG: [RememberanceAgentService:getDocument("+docid+")] warning! did not reat specified number of bytes");}
//     Logger.debug.println("DEBUG: [RememberanceAgentService:getDocument("+docid+")] I reat \""+(new String(retval))+"\"");

     aval = err.available();
     if(aval>0)
       {
        byte retvalerr[] = new byte[aval];
        err.read(retvalerr);

        Logger.err.println("ERROR: [RememberanceAgentService] Maybe I cannot read "+(new String(retvalerr)));
       };
     return(new String(retval));
    }
  catch(Exception e)
    {
     e.printStackTrace(Logger.err);
     return(null);
    }

  }


  }
